m = 809
n = euler_phi(m)
K.<X> = CyclotomicField(m)
p = m
P = 1-X

error_param = 5

### defining manual reduction mod p, as sage is pedantic about it
def modp(h):
    f = K(0)
    for i in range(n):
	f+= (h[i] % p) * X^i
    return f


### defining manual reduction mod P, as sage is pedantic about it
def modP(h):
    return sum(h) % p


############## The RLWE oracle

### Sampling according to the trick of [DucasDurmus]
### Temporary Hack: usding uniform errors, but this 
### should still be meaningful

def tern_error():
	e = K(0)
	for i in range(m):
		e += randint(-error_param,error_param) * X^i
	return e

def rand_poly():
	e = K(0)
	for i in range(n):
		e += randint(0,p-1) * X^i
	return e

def RLWE(s):
	a = rand_poly()
	e = tern_error()
	b = modp(a*s+e)
	return (a,b)


s = rand_poly()
def RLWE_s():
	return RLWE(s)



############# The distinguisher

### This distinguisher should be optimal (reaching the bounds given by the KL-divergence)
### see [Vaudenay,  Decorrelation: A theory for block cipher security]

### Compute exactly the distribution of e mod P as a probability density table
### To do so, start with distribution that is identically 0, and convolve it
### successively with the terms of the error reduced mod P 
###
### This relies on the identity : e mod P = \sum e_i \bmod p

pdf = p*[0.]
pdf[0] = 1.

dom_ei = range(error_param+1)+range(p - error_param,p)
dom_em1 = [2*x % p for x in dom_ei]
f = 1./len(dom_ei)



pdf_ei = p*[0.]
for x in dom_ei:
	pdf_ei[x] = f

pdf_em1 = p*[0.]
for x in dom_em1:
	pdf_em1[x] = f


def convol(a,b, domb):
    c = p*[0.]
    for i in range(p):
        for j in domb:
            c[(i+j) % p] += a[i] * b[j]
    return c

print "build pdf_table"

for i in range(m-1):
    pdf = convol(pdf,pdf_ei,dom_ei)

  
pdf = convol(pdf,pdf_em1,dom_em1)

print "This should be ~1 :", sum(pdf)
draw = line([(i,pdf[i]) for i in range(p)])
draw.save("pdf.png")
print "Probability density distrib plotted to pdf.png"

def scoring(samples):
    sc = 1.
    for x in samples:
        sc*= pdf[x]
    return sc


############# The Ramification attack 

goal =  modP(s)
print "The right value of s modP to be found:", goal
print "sanity check: modP is a ring morphism ?", modP(s*s) == goal*goal %p

nbsample = 20
RLWE_sample = [RLWE_s() for i in range(nbsample)]

def guess_s(samples):
    best_guess = 0
    best_score = 0.
    sample_ModP = [(modP(a), modP(b)) for (a,b) in samples ]  
    for g in range(p):
        L = [(rb - g*ra) % p for (ra,rb) in sample_ModP]
        sc = scoring(L)
        if sc>best_score:
            best_guess = g
            best_score = sc
    return best_guess

print "best guess : ", guess_s(RLWE_sample)